#include <l4/sys/types.h>
#include <l4/sys/syscalls.h>
#include <l4/sys/ipc.h>
#include <l4/sys/kernel.h>
#include <l4/sys/kdebug.h>
#include <l4/rmgr/librmgr.h> 

#include <stdio.h>
#include <assert.h>

#define MEM_SIZE 0x01000000
#define MAP_ADDR 0x40000000

extern l4_threadid_t rmgr_pager_id;

/*****************************************************************************/
/**
 * \brief  Request a specific page.
 * 
 * \param  page          Page address
 *	
 * \return 0 on success, -1 if mapping failed.
 */
/*****************************************************************************/ 
static int
sigma0_map_page(l4_umword_t page)
{
  int error;
  l4_umword_t base;
  l4_fpage_t fpage;
  l4_msgdope_t result;

  /* call sigma0 to map the page */
  page &= L4_PAGEMASK;
  error = l4_i386_ipc_call(rmgr_pager_id,L4_IPC_SHORT_MSG,page,0,
			   L4_IPC_MAPMSG(MAP_ADDR + page, 
					 L4_LOG2_PAGESIZE),
			   &base,&fpage.fpage,L4_IPC_NEVER,&result);
  if (error)
    {
      printf("calling sigma0 failed (result 0x%08x)!\n",result.msgdope);
      return -1;
    }

  if ((fpage.fpage == 0) || (!l4_ipc_fpage_received(result)))
    /* sigma0 denied page */
    return -1;

  /* done */
  return 0;
}

/*****************************************************************************/
/**
 * \brief  Request a specific 4M-page
 * 
 * \param  page          Page address
 *	
 * \return 0 on success, -1 if mapping failed.
 */
/*****************************************************************************/ 
static int
sigma0_map_4Mpage(l4_umword_t page)
{
  int error;
  l4_umword_t base;
  l4_fpage_t fpage;
  l4_msgdope_t result;

  /* call sigma0 to map the page */
  page &= L4_SUPERPAGEMASK;
  error = l4_i386_ipc_call(rmgr_pager_id,L4_IPC_SHORT_MSG,
			   page | 1, L4_LOG2_SUPERPAGESIZE << 2,
			   L4_IPC_MAPMSG(MAP_ADDR + page,
					 L4_LOG2_SUPERPAGESIZE),
			   &base,&fpage.fpage,L4_IPC_NEVER,&result);

  if (error)
    {
      printf("calling sigma0 failed (result 0x%08x)!\n",result.msgdope);
      return -1;
    }

  if ((fpage.fpage == 0) || (!l4_ipc_fpage_received(result)))
    /* sigma0 denied page */
    return -1;

  if (fpage.fp.size != L4_LOG2_SUPERPAGESIZE)
    {
      /* no 4M-page received, this can happen if the whole 4M-page is not
       * available but the 4K-page at that address */

      /* unmap page, the 4K-page must be mapped explicitly */
      fpage.fp.page += (MAP_ADDR >> L4_LOG2_PAGESIZE);
      l4_fpage_unmap(fpage,L4_FP_FLUSH_PAGE | L4_FP_ALL_SPACES);

      return -1;
    }

  /* done */
  return 0;
}

/*****************************************************************************/
/**
 * \brief Main.
 */
/*****************************************************************************/ 
int 
main(void)
{
  l4_umword_t start_addr,end_addr,addr;
  int ret;

  printf ("rmgrtest running\n");

  enter_kdebug("begin.");

  rmgr_init();

  /* reserve memory atr rmgr */
  start_addr = rmgr_reserve_mem(MEM_SIZE,0,0,0,0);
  assert(start_addr != -1);
  end_addr = start_addr + MEM_SIZE;

  printf("got memory area 0x%08x - 0x%08x\n",start_addr,start_addr + MEM_SIZE);
  
  /* map */
  addr = start_addr;
   while (addr < end_addr)
    {
      if (((addr + L4_SUPERPAGESIZE) <= end_addr) && 
	  ((addr & ~L4_SUPERPAGEMASK) == 0))
	{
	  /* try to get 4M page */
	  ret = sigma0_map_4Mpage(addr);
	  if (ret == 0)
	    {
	      addr += L4_SUPERPAGESIZE;
	      continue;
	    }
	}
 
      /* map 4K page */
      sigma0_map_page(addr);
      
      addr += L4_PAGESIZE;
    }

  enter_kdebug("map done.");

  return 0;
}
