#include <stdio.h>
#include <unistd.h>

#include <flux/machine/pc/direct_cons.h>
#include <flux/machine/pc/reset.h>
#include <flux/machine/pio.h>
#include <flux/c/string.h>
#include <flux/base_critical.h>

#include <l4/sys/kdebug.h>

#include "serial.h"

extern int hercules;
extern int kernel_running;

static void herc_direct_cons_putchar(unsigned char c);
static void kern_putchar(unsigned char c);
static int direct_cons_try_getchar(void);
int have_hercules(void);
void my_pc_reset(void);

int
putchar(int c)
{
  if (kernel_running)
    kern_putchar(c);
  else
    {
      if (hercules)
	herc_direct_cons_putchar(c);
      else
	direct_cons_putchar(c);
      com_cons_putchar(c);
    }

  return c;
}

int
getchar(void)
{
  if (kernel_running)
    return direct_cons_getchar();
  else
    {
      int c;
      do
	{
	  c = com_cons_try_getchar();
	  if (c == -1)
	    c = direct_cons_try_getchar();
	} while (c == -1);
      return c;
    }
}

void
my_pc_reset(void)
{
  outb_p(0x70, 0x80);
  inb_p (0x71);

  while (inb(0x64) & 0x02)
    ;

  outb_p(0x70, 0x8F);
  outb_p(0x71, 0x00);

  outb_p(0x64, 0xFE);

  for (;;)
    ;
}

void 
_exit(int fd)
{
  char c;

  if (kernel_running)
    {
      printf("\nReturn reboots, \"k\" enters L4 kernel debugger...\n");
      c = getchar();
      if (c == 'k' || c == 'K')
	enter_kdebug("before reboot");
    }
  else
    {
      printf("\nReturn reboots...\n");
      getchar();
    }

  my_pc_reset();
}

/* modified direct_cons_putchar from oskit/libkern */
#define HERC_VIDBASE ((unsigned char *)0xb0000)
void
herc_direct_cons_putchar(unsigned char c)
{
	static int ofs = -1;

	base_critical_enter();

	if (ofs < 0)
	{
		/* Called for the first time - initialize.  */
		ofs = 0;
		herc_direct_cons_putchar('\n');
	}

	switch (c)
	{
	case '\n':
		bcopy(HERC_VIDBASE+80*2, HERC_VIDBASE, 80*2*24);
		bzero(HERC_VIDBASE+80*2*24, 80*2);
		/* fall through... */
	case '\r':
		ofs = 0;
		break;

	case '\t':
		ofs = (ofs + 8) & ~7;
		break;

	default:
		/* Wrap if we reach the end of a line.  */
		if (ofs >= 80) {
			herc_direct_cons_putchar('\n');
		}

		/* Stuff the character into the video buffer. */
		{
			volatile unsigned char *p = HERC_VIDBASE
			  			  + 80*2*24 + ofs*2;
			p[0] = c;
			p[1] = 0x0f;
			ofs++;
		}
		break;
	}

	base_critical_leave();
}

static void
kern_putchar(unsigned char c)
{
  outchar(c);
  if (c == '\n')
    outchar('\r');
}

int
have_hercules(void)
{
  unsigned short *p, p_save;
  int count = 0;
  unsigned long delay;
  
  /* check for video memory */
  p = (unsigned short*)0xb0000;
  p_save = *p;
  *p = 0xaa55; if (*p == 0xaa55) count++;
  *p = 0x55aa; if (*p == 0x55aa) count++;
  *p = p_save;
  if (count != 2)
    return 0;

  /* check for I/O ports (HW cursor) */
  outb_p(0x3b4, 0x0f);
  outb  (0x3b5, 0x66);
  for (delay=0; delay<0x100000UL; delay++)
    asm volatile ("nop" : :);
  if (inb_p(0x3b5) != 0x66)
    return 0;
  
  outb_p(0x3b4, 0x0f);
  outb  (0x3b5, 0x99);
  for (delay=0; delay<0x100000UL; delay++)
    asm volatile ("nop" : :);
  if (inb_p(0x3b5) != 0x99)
    return 0;

  /* reset position */
  outb_p(0x3b4, 0x0f);
  outb  (0x3b5, 0x0);

  return 1;
}

#include <flux/x86/pio.h>
#include <flux/x86/pc/keyboard.h>

#define SHIFT -1

static const char keymap[128][2] = {
  {0},			/* 0 */
  {27,	27},		/* 1 - ESC */
  {'1',	'!'},		/* 2 */
  {'2',	'@'},
  {'3',	'#'},
  {'4',	'$'},
  {'5',	'%'},
  {'6',	'^'},
  {'7',	'&'},
  {'8',	'*'},
  {'9',	'('},
  {'0',	')'},
  {'-',	'_'},
  {'=',	'+'},
  {8,	8},		/* 14 - Backspace */
  {'\t',	'\t'},		/* 15 */
  {'q',	'Q'},
  {'w',	'W'},
  {'e',	'E'},
  {'r',	'R'},
  {'t',	'T'},
  {'y',	'Y'},
  {'u',	'U'},
  {'i',	'I'},
  {'o',	'O'},
  {'p',	'P'},
  {'[',	'{'},
//   {']',	'}'},		/* 27 */
  {'+',	'*'},		/* 27 */
  {'\r',	'\r'},		/* 28 - Enter */
  {0,	0},		/* 29 - Ctrl */
  {'a',	'A'},		/* 30 */
  {'s',	'S'},
  {'d',	'D'},
  {'f',	'F'},
  {'g',	'G'},
  {'h',	'H'},
  {'j',	'J'},
  {'k',	'K'},
  {'l',	'L'},
  {';',	':'},
  {'\'',	'"'},		/* 40 */
  {'`',	'~'},		/* 41 */
  {SHIFT,	SHIFT},		/* 42 - Left Shift */
  {'\\',	'|'},		/* 43 */
  {'z',	'Z'},		/* 44 */
  {'x',	'X'},
  {'c',	'C'},
  {'v',	'V'},
  {'b',	'B'},
  {'n',	'N'},
  {'m',	'M'},
  {',',	'<'},
  {'.',	'>'},
  //  {'/',	'?'},		/* 53 */
  {'-',	'_'},		/* 53 */
  {SHIFT,	SHIFT},		/* 54 - Right Shift */
  {0,	0},		/* 55 - Print Screen */
  {0,	0},		/* 56 - Alt */
  {' ',	' '},		/* 57 - Space bar */
  {0,	0},		/* 58 - Caps Lock */
  {0,	0},		/* 59 - F1 */
  {0,	0},		/* 60 - F2 */
  {0,	0},		/* 61 - F3 */
  {0,	0},		/* 62 - F4 */
  {0,	0},		/* 63 - F5 */
  {0,	0},		/* 64 - F6 */
  {0,	0},		/* 65 - F7 */
  {0,	0},		/* 66 - F8 */
  {0,	0},		/* 67 - F9 */
  {0,	0},		/* 68 - F10 */
  {0,	0},		/* 69 - Num Lock */
  {0,	0},		/* 70 - Scroll Lock */
  {'7',	'7'},		/* 71 - Numeric keypad 7 */
  {'8',	'8'},		/* 72 - Numeric keypad 8 */
  {'9',	'9'},		/* 73 - Numeric keypad 9 */
  {'-',	'-'},		/* 74 - Numeric keypad '-' */
  {'4',	'4'},		/* 75 - Numeric keypad 4 */
  {'5',	'5'},		/* 76 - Numeric keypad 5 */
  {'6',	'6'},		/* 77 - Numeric keypad 6 */
  {'+',	'+'},		/* 78 - Numeric keypad '+' */
  {'1',	'1'},		/* 79 - Numeric keypad 1 */
  {'2',	'2'},		/* 80 - Numeric keypad 2 */
  {'3',	'3'},		/* 81 - Numeric keypad 3 */
  {'0',	'0'},		/* 82 - Numeric keypad 0 */
  {'.',	'.'},		/* 83 - Numeric keypad '.' */
};

int
direct_cons_try_getchar(void)
{
  static unsigned shift_state;
  unsigned status, scan_code, ch;

  base_critical_enter();

retry:

  /* Wait until a scan code is ready and read it. */
  status = inb(0x64);
  if ((status & K_OBUF_FUL) == 0) {
    base_critical_leave();
    return -1;
  }
  scan_code = inb(0x60);

  /* Drop mouse events */
  if ((status & K_AUX_OBUF_FUL) != 0) {
    base_critical_leave();
    return -1;
  }

  /* Handle key releases - only release of SHIFT is important. */
  if (scan_code & 0x80) {
    scan_code &= 0x7f;
    if (keymap[scan_code][0] == SHIFT)
      shift_state = 0;
    goto retry;
  }

  /* Translate the character through the keymap. */
  ch = keymap[scan_code][shift_state];
  if (ch == (unsigned)SHIFT) {
    shift_state = 1;
    goto retry;
  } else if (ch == 0)
    goto retry;

  base_critical_leave();

  return ch;
}

int
direct_cons_getchar(void)
{
  int c;

  do
    {
      c = direct_cons_try_getchar();
    } while (c == -1);

  return c;
}

